//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#include "Utils.hpp"
#include "SoundManager.hpp"

#include <cmath>
#include <iostream>

#include "MyException.hpp"
#include <SFML/Graphics.hpp>
//------------------------------------------------------------------------------
using namespace std;
//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
Vector3D Vector3D::operator +(const Vector3D& vect) const
{
    return sf::operator+(*this, vect);
}
//------------------------------------------------------------------------------
Vector3D Vector3D::operator -(const Vector3D& vect) const
{
    return sf::operator-(*this, vect);
}
//------------------------------------------------------------------------------
bool Vector3D::operator ==(const Vector3D& vect) const
{
    return sf::operator==(*this, vect);
}
//------------------------------------------------------------------------------
Vector3D Vector3D::operator *(float nb) const
{
    return sf::operator*(*this, nb);
}
//------------------------------------------------------------------------------
Vector3D Vector3D::operator /(float nb) const
{
    return sf::operator/(*this, nb);
}
//------------------------------------------------------------------------------
/** Get the sum of the squares of x and y (basicaly, x² + y² + z²). This is useful to compare the vectors size without using a square root (which is more time consuming). */
float Vector3D::GetSquareSum() const
{
    return x * x + y * y + z * z;
}
//------------------------------------------------------------------------------

/** Get the size of the Vector */
float Vector3D::GetSize() const
{
    return sqrt(GetSquareSum());
}
//------------------------------------------------------------------------------

/** Get the normalised vector (direction). */
Vector3D Vector3D::GetDirection() const
{
    return *this / GetSize();
}
//------------------------------------------------------------------------------

/** Get the dot product / scalar product of this vector with another one */
float Vector3D::GetDotProduct(const Vector3D &vector2) const
{
    return x * vector2.x + y * vector2.y + z * vector2.z;
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
Vector2D PointRotate(const Vector2D &centre, const Vector2D &pt, float angle)
{
    Vector2D newPt;
    float cosAngle = cos(angle);
    float sinAngle = sin(angle);

    newPt.x = ((pt.x - centre.x) * cosAngle) - ((centre.y - pt.y) * sinAngle);
    newPt.y = ((pt.y - centre.y) * cosAngle) - ((pt.x - centre.x) * sinAngle);
    return newPt;
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
Vector2D Vector2D::operator +(const Vector2D& vect) const
{
    return sf::operator+(*this, vect);
}
//------------------------------------------------------------------------------
Vector2D Vector2D::operator -(const Vector2D& vect) const
{
    return sf::operator-(*this, vect);
}
//------------------------------------------------------------------------------
bool Vector2D::operator ==(const Vector2D& vect) const
{
    return sf::operator==(*this, vect);
}
//------------------------------------------------------------------------------
Vector2D Vector2D::operator *(float nb) const
{
    return sf::operator*(*this, nb);
}
//------------------------------------------------------------------------------
Vector2D Vector2D::operator /(float nb) const
{
    return sf::operator/(*this, nb);
}
//------------------------------------------------------------------------------
/** Get the sum of the squares of x and y (basicaly, x² + y²). This is useful to compare the vectors size without using a square root (which is more time consuming). */
float Vector2D::GetSquareSum() const
{
    return x * x + y * y;
}
//------------------------------------------------------------------------------

/** Get the size of the Vector */
float Vector2D::GetSize() const
{
    return sqrt(GetSquareSum());
}
//------------------------------------------------------------------------------

/** Get the normalised vector (direction). */
Vector2D Vector2D::GetDirection() const
{
    return *this / GetSize();
}
//------------------------------------------------------------------------------

/** Get the dot product / scalar product of this vector with another one */
float Vector2D::GetDotProduct(const Vector2D &vector2) const
{
    return x * vector2.x + y * vector2.y;
}
//------------------------------------------------------------------------------

/** Get the angle between this vector and the x axis */
float Vector2D::GetAngle() const
{
    if(y == 0 && x < 0)
    {
        return M_PI;
    }

    return - 2 * atan(y / (x + GetSize())); // Cf http://fr.wikipedia.org/wiki/Coordonn%C3%A9es_polaires
}
//------------------------------------------------------------------------------
/** Get the normal to the vector, in the <b>direct</b> direction (<b>+</b> pi/2). The returned coordonates are Vector2D(-y, x). */
Vector2D Vector2D::GetDirectNormal() const
{
    return Vector2D(-y, x);
}
//------------------------------------------------------------------------------
/** Get the normal to the vector, in the <b>indirect</b> direction (<b>-</b> pi/2). The returned coordonates are Vector2D(y, -x). */
Vector2D Vector2D::GetIndirectNormal() const
{
    return Vector2D(y, -x);
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
/** Constructor to create an empty sound */
Sound::Sound() : sf::SoundBuffer(), ClassAllocCounter()
{

}
//------------------------------------------------------------------------------
/** Create a sound and load it from a file */
Sound::Sound(const std::string &filename) : sf::SoundBuffer(), ClassAllocCounter()
{
    LoadFromFile(filename);
}
//------------------------------------------------------------------------------
Sound::~Sound()
{
    MakeSoundBufferCallback();
}
//------------------------------------------------------------------------------
void Sound::LoadFromFile(const std::string &filename)
{
    if(!sf::SoundBuffer::LoadFromFile(filename))
    {
        MY_THROW("Error: Can't load sound '" << filename << "' !");
    }
}
//------------------------------------------------------------------------------
/** Get the sample rate */
int Sound::GetSampleRate() const
{
    return sf::SoundBuffer::GetSampleRate();
}
//------------------------------------------------------------------------------

/** Return the number of channels (1 = mono, 2 = stereo, ...) */
int Sound::GetChannelsCount() const
{
    return sf::SoundBuffer::GetChannelsCount();
}
//------------------------------------------------------------------------------
void Sound::AddSoundBufferCallback(sndBufferPtr ptr)
{
    m_callbackSet.insert(ptr);
}
//------------------------------------------------------------------------------
void Sound::RemoveSoundBufferCallback(sndBufferPtr ptr)
{
    m_callbackSet.erase(ptr);
}
//------------------------------------------------------------------------------
void Sound::MakeSoundBufferCallback()
{
    if(!m_callbackSet.empty())
    {
        cerr << "Warning: You are deleting a sound used by a Sound Player !" << endl;

        for(std::set<sndBufferPtr>::iterator it = m_callbackSet.begin(); it != m_callbackSet.end(); it++)
        {
            (*it)->Stop();
        }
        m_callbackSet.clear();
    }
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
void Texture::LoadFromFile(const std::string &filename)
{
    if(!sf::Image::LoadFromFile(filename))
    {
        MY_THROW("Error: Can't load texture '" << filename << "' !");
    }
}
//------------------------------------------------------------------------------
Texture::Texture() : sf::Image(), ClassAllocCounter()
{

}
//------------------------------------------------------------------------------
Texture::Texture(const std::string &filename) : sf::Image(), ClassAllocCounter()
{
    LoadFromFile(filename);
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
/** Constructor to create an empty music */
Music::Music() : sf::Music(), ClassAllocCounter()
{

}
//------------------------------------------------------------------------------
/** Create a music and load it from a file */
Music::Music(const std::string &filename) : sf::Music(), ClassAllocCounter()
{
    LoadFromFile(filename);
}
//------------------------------------------------------------------------------
void Music::LoadFromFile(const std::string &filename)
{
    if(!sf::Music::OpenFromFile(filename))
    {
        MY_THROW("Error: Can't load music '" << filename << "' !");
    }
}
//------------------------------------------------------------------------------
bool Music::IsPlaying() const
{
    return sf::Music::GetStatus() == sf::Music::Playing;
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
/** Constructor to create an empty font */
Font::Font() : sf::Font(), ClassAllocCounter()
{

}
//------------------------------------------------------------------------------
/** Create a font and load it from a file */
Font::Font(const std::string &filename) : sf::Font(), ClassAllocCounter()
{
    LoadFromFile(filename);
}
//------------------------------------------------------------------------------
void Font::LoadFromFile(const std::string &filename)
{
    if(!sf::Font::LoadFromFile(filename))
    {
        MY_THROW("Error: Can't load font '" << filename << "' !");
    }
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
sf::Unicode::UTF8String StrToSfUtf8(const std::string &str)
{
	return sf::Unicode::UTF8String(reinterpret_cast<const unsigned char*>(str.c_str()));
}
//------------------------------------------------------------------------------
} // namespace En
